# version
__version__ = "0.1.5"
# author
__author__ = "Sina Gilassi"
# email
__email__ = "sina.gilassi@gmail.com"
# license
__license__ = "MIT"
# description
__description__ = "PyThermoFlash is a Python package for performing thermodynamic flash calculations and property estimations for various fluid systems."
